import 'dart:async';
import 'dart:io';
import 'package:heavylist/heavylist.dart';
import 'package:jsonstreamreader/abstractreader.dart';

import 'complexexpression.dart';
import 'jsontoken.dart';
import 'jsontokenizer.dart';
import 'source/remote/remotesource.dart';
import 'state/globalreadermanager.dart';
import 'state/globalreaderstatemanager.dart';
import 'state/reader/readerstate.dart';
import 'streamer.dart';

///Reads a List of JsonTokens and calls the relevant callbacks
class Reader extends AbstractReader {
  JsonTokenizer tokenizer;
  Function _trailing;
  List<MapEntry<ComplexExpression, Function>> _filters;
  Function _done;
  Function _fail;
  Function _progress;
  Function _onpause;
  Duration _delay;
  int _combine = 0;
  ReaderState _state;
  bool _paused = false;
  bool _running = false;
  bool _hault = false;
  String _reference = 'N';
  HeavyList<MapEntry<String, dynamic>> _list;
  StreamSubscription<List<MapEntry<String, dynamic>>> _subscribe;
  StreamController _parsing = new StreamController();

  ///Create a new Reader from ReaderState
  Reader.fromState(ReaderState state) {
    GlobalReaderManager.addReader(this);
    _reference = 'S';
    this._delay = new Duration(microseconds: state.delay);
    this.tokenizer = new JsonTokenizer.fromState(
        new Streamer(new File('/empty')), state.group,
        combine: 1);
    //recreate JsonTokens
    JsonToken.fromState(state.group, state.jsontokenstate);
    _state = state;
    new Timer(new Duration(microseconds: 0), () {
      if (!_hault) {
        _stateFire();
      }
    });
  }
  void _stateFire() {
    this._parse(_state.parse).then((String d) {
      new Timer(_delay, () {
        this._partial(_state);
      });
    }).catchError((err) {
      if (this._fail != null) {
        _parsing.close();
        this._fail(err);
      }
    });
  }

  void _partial(ReaderState state) {
    Streamer stream;
    new Timer(new Duration(microseconds: 1), () {
      //parse state.parse if>0 and then run below
      if (state.isRemote) {
        RemoteSource r = new RemoteSource.fromString(state.file);
        stream = new Streamer(r,
            start: state.start, end: state.end, chunksize: state.chunksize);
        new Timer(new Duration(microseconds: 0), () {
          r.download();
        });
      } else {
        File r = new File(state.file);
        stream = new Streamer(r,
            start: state.start, end: state.end, chunksize: state.chunksize);
      }
      this._combine = state.combine;
      this._delay = new Duration(microseconds: state.delay);
      this._filters = new List<MapEntry<ComplexExpression, Function>>();
      this.tokenizer = new JsonTokenizer.fromState(stream, state.group,
          combine: this._combine);

      new Timer(new Duration(microseconds: 0), () {
        this._run();
      });
    });
  }

  ///@stream Steamer
  ///int delay how long should the reader wait before reading current and next sequence
  ///combine 1 converts idenified array values with the same parent into an object
  ///combine 2 uses the first common parent
  Reader(Streamer stream, {int delay = 0, combine = 0}) {
    GlobalReaderManager.addReader(this);
    this._combine = combine;
    this._delay = new Duration(microseconds: delay);
    this._filters = new List<MapEntry<ComplexExpression, Function>>();
    this.tokenizer = new JsonTokenizer(stream, combine: this._combine);

    new Timer(new Duration(microseconds: 0), () {
      if (!_hault) {
        _fire();
      }
    });
  }
  void _fire() {
    new Timer(new Duration(microseconds: 0), () {
      this._run();
    });
  }

  _run() {
    this.tokenizer.tokenize().then((JsonResponse item) {
      if (item.type == 0) {
        this._parse(item.data).then((d) {
          _run();
        }).catchError((e) {
          if (this._fail != null) {
            _parsing.close();
            this._fail(e);
          }
        });
      } else {
        if (this._done != null) {
          _parsing.close();
          this._done();
        }
      }
    }).catchError((err) {
      if (this._fail != null) {
        _parsing.close();
        this._fail(err);
      }
    });
  }

  Future<String> _parse(List<MapEntry<String, dynamic>> parse) async {
    this.tokenizer.destroy();
    Completer<String> completer = new Completer<String>();
    this._list = new HeavyList<MapEntry<String, dynamic>>(parse);
    _parsing.add(null);
    _running = true;
    _list.loop(new Duration(microseconds: 0),
        (List<MapEntry<String, dynamic>> origin) {
      _list.empty();
      if (this._progress != null) {
        this._progress(Streamer.progess(this.tokenizer.identity));
      }
      completer.complete("done");
    }, (MapEntry<String, dynamic> item, Function resume) {
      dynamic firstValue = item.value;
      dynamic firstKey = item.key;
      Iterable<String> ikeys = [];
      List<Function> functions = [];
      if (firstValue.runtimeType.toString() ==
          '_InternalLinkedHashMap<String, dynamic>') {
        ikeys = item.value.keys;
      }
      if (this._trailing != null) {
        functions.add(this._trailing);
      }
      this._filters.forEach((MapEntry<ComplexExpression, Function> filter) {
        if (filter.key.evaluate(ikeys, firstKey)) {
          functions.add(filter.value);
        }
      });
      if (functions.length > 0) {
        Future.forEach(functions, (Function func) {
          return func(firstValue, firstKey);
        }).then((d) {
          this._list.clearRegistered();
          resume();
        });
      } else {
        this._list.clearRegistered();
        resume();
      }
    });
    return completer.future;
  }

  ///return all objects or values found
  Reader trailing(dynamic func(dynamic value, String key)) {
    this._trailing = func;
    return this;
  }

  ///called when reader is complete
  Reader done(void func()) {
    this._done = func;
    return this;
  }

  ///called when reader fails
  Reader fail(void func(dynamic err)) {
    this._fail = func;
    return this;
  }

  ///double progess of reader
  Reader progress(void func(double progress)) {
    this._progress = func;
    return this;
  }

  ///Prevent the reader from starting immediately
  void hault() {
    _hault = true;
  }

  ///if the reader has hault,start now
  void begin() {
    if (_hault) {
      _hault = false;
      if (this._reference == 'N') {
        _fire();
      } else {
        _stateFire();
      }
    }
  }

  ///Filter allows users to filter object or values by JsonPath(eg $.items.0) or PathExpression(eg. /items/0/)
  Reader filter(
      String expression, Future<dynamic> func(dynamic value, String key)) {
    this._filters.add(new MapEntry(new ComplexExpression(expression), func));
    return this;
  }

  ///remove filter at index
  void removeFilter(int index) {
    this._filters.removeAt(0);
  }

  ///called everytime reader is paused
  Reader onpause(void func(ReaderState state)) {
    this._onpause = func;
    return this;
  }

  ///Return the State Reader was in upon pausing
  Future<ReaderState> pause() async {
    Completer<ReaderState> completer = new Completer<ReaderState>();
    if (!this._paused) {
      if (_running) {
        new Timer(new Duration(microseconds: 0), () {
          this._pause();
          completer.complete(this._state);
        });
      } else {
        StreamSubscription sb;
        sb = _parsing.stream.listen((onData) {
          this._pause();
          sb.cancel();
          if (!completer.isCompleted) {
            completer.complete(this._state);
          }
        });
      }
    }
    return completer.future;
  }

  void _pause() {
    this._paused = true;
    _subscribe.pause();
    this._list.pause();
    this._state = new ReaderState(
        JsonToken.getState(this.tokenizer.identity),
        this.tokenizer.identity,
        this.tokenizer.stream.path,
        this.tokenizer.stream.start,
        this.tokenizer.stream.end,
        this._delay.inMicroseconds,
        this._combine,
        this.tokenizer.stream.chunksize,
        this._list.getList());
    if (this._onpause != null) {
      this._onpause(this._state);
    }
    GlobalReaderStateManager.addState(this._state);
  }

  ///continue reader
  void resume() {
    if (this._paused) {
      this._paused = false;
      this._list.resume();
      _subscribe.resume();
      GlobalReaderStateManager.removeState(this._state);
    }
  }
}
