import 'dart:async';
import 'dart:convert';
import 'simpletoken.dart';

import 'state/filestate.dart';

import 'source/normalized.dart';
import 'streamerunit.dart';

///Identified chunk for a file based on
///chunksize
class Streamer {
  //chunk size
  int _size;
  //max size in kb of each chunk
  dynamic _max;
  int _end;
  Normalized _file;
  static Map<int, double> _progress = new Map<int, double>();
  int group;
  int _multiplier = 1000;
  bool _findLength = false;
  StreamTransformer _transformer;

  ///
  /// @file file to be parsed
  /// @start where to start parsing
  /// @end where to stop parsing
  /// @chunksize can be an int or a List,eg [10,100] first collect the first 10KB and 100KB everytime after
  /// @unit in what unit is the chunksize represented Bytes,Kilobytes or Megabytes
  /// @findLength here the streamer won't identify the size of the file first
  ///
  Streamer(var file,
      {start = 0,
      end = 0,
      dynamic chunksize = 100,
      bool findLength = false,
      STREAMERUNIT unit = STREAMERUNIT.KILOBYTES}) {
    String runtime = file.runtimeType.toString();
    this._findLength = findLength;
    if (runtime != '_File' && runtime != 'RemoteSource') {
      throw "Streamer accepts only File or RemoteSource";
    }
    if (unit == STREAMERUNIT.BYTES) {
      this._multiplier = 1;
    } else if (unit == STREAMERUNIT.MEGABYTES) {
      this._multiplier = 1000000;
    }
    this._file = new Normalized(file);
    this._end = end;
    this._file.setBoundary(start, end);
    this._size = 0;
    this._transformer =
        new StreamTransformer<String, List<SimpleToken>>.fromHandlers(
            handleData: _handleData);
    FileState.addState(file);
    this._max = chunksize;
  }

  ///set the length of the file
  void setLength(int length) {
    this._file.setLength(length);
  }

  void _handleData(String data, EventSink<List<SimpleToken>> sink) {
    sink.add(this.tidy(data));
  }

  List<SimpleToken> tidy(String sequence) {
    bool lastisValue = false;
    bool previouslastisValue = false;
    List<SimpleToken> parts = [];
    int type = SimpleToken.NULL;
    String previous = sequence[0]
        .replaceAll(new RegExp(r"[\r\n\t]+(.+[\r\n\t]+.+)[\t\r\n]+"), '');
    if (previous == ",") {
      type = SimpleToken.ARRAYSEPERATOR;
    } else if (previous == ':') {
      type = SimpleToken.OBJECTSEPERATOR;
    } else if (previous == '{') {
      type = SimpleToken.OBJECTSTART;
    } else if (previous == '}') {
      type = SimpleToken.OBJECTEND;
    } else if (previous == '[') {
      type = SimpleToken.ARRAYSTART;
    } else if (previous == ']') {
      type = SimpleToken.ARRAYEND;
    } else if (previous == '"') {
      type = SimpleToken.VALUE;
    } else {
      type = SimpleToken.VALUE;
      lastisValue = true;
    }
    parts.add(new SimpleToken(previous, type));
    //no need for .replaceAll(new RegExp(r'(\\)+"'), '&nbqts;')
    int len = sequence.length;
    for (int i = 1; i < len; i++) {
      String current = sequence[i];
      if (current == '\n' || current == '\t' || current == '\r') {
        continue;
      }
      if (current == '"') {
        parts.add(new SimpleToken(current, SimpleToken.VALUE));
        previouslastisValue = lastisValue;
        lastisValue = false;
      } else if (current == ',') {
        parts.add(new SimpleToken(current, SimpleToken.ARRAYSEPERATOR));
        previouslastisValue = lastisValue;
        lastisValue = false;
      } else if (current == ':') {
        parts.add(new SimpleToken(current, SimpleToken.OBJECTSEPERATOR));
        previouslastisValue = lastisValue;
        lastisValue = false;
      } else if (current == '{') {
        parts.add(new SimpleToken(current, SimpleToken.OBJECTSTART));
        previouslastisValue = lastisValue;
        lastisValue = false;
      } else if (current == '}') {
        parts.add(new SimpleToken(current, SimpleToken.OBJECTEND));
        previouslastisValue = lastisValue;
        lastisValue = false;
      } else if (current == '[') {
        parts.add(new SimpleToken(current, SimpleToken.ARRAYSTART));
        previouslastisValue = lastisValue;
        lastisValue = false;
      } else if (current == ']') {
        parts.add(new SimpleToken(current, SimpleToken.ARRAYEND));
        previouslastisValue = lastisValue;
        lastisValue = false;
      } else if (!lastisValue && current != ' ') {
        parts.add(new SimpleToken(current, SimpleToken.VALUE));
        previouslastisValue = lastisValue;
        lastisValue = true;
      } else if (!lastisValue && current == ' ') {
        lastisValue = false;
      } else {
        parts[parts.length - 1].value += current;
        previouslastisValue = lastisValue;
        lastisValue = true;
      }
      if (!previouslastisValue && lastisValue && i > 1) {
        int index = parts.length - 1;
        parts[index].value = parts[index].value.trimLeft();
      } else if (previouslastisValue && !lastisValue && i > 1) {
        int index = parts.length - 2;
        parts[index].value = parts[index].value.trimRight();
      }
    }
    return parts;
  }

  setGroup(int group) {
    this.group = group;
    Streamer._progress[this.group] = 0.0;
  }

  get end {
    return this._end;
  }

  get start {
    return this._file.getLast();
  }

  dynamic get file {
    return this._file.file;
  }

  dynamic get path {
    return this._file.path;
  }

  String get normalized {
    return this._file.file;
  }

  int get chunksize {
    return this._max;
  }

  Stream<List<SimpleToken>> next() {
    int currentMax = 100;
    if (this._max.runtimeType.toString().contains('List')) {
      currentMax = this._max.length > 1 ? this._max.removeAt(0) : this._max[0];
    } else {
      currentMax = this._max;
    }
    return this
        ._file
        .next(currentMax * this._multiplier, findLength: this._findLength)
        .transform(utf8.decoder)
        .transform(_transformer);
  }

  setProgress() {
    Streamer._progress[this.group] = this._file.getProgress() * 100;
  }

  static double progess(group) {
    return Streamer._progress[group];
  }
}
